#include <string.h>
#include "globals.h"
#include "arm.h"
#include "uart.h"
#include "timer.h"
#include "num2str.h"
#include "ui.h"
#include "block.h" /* strcasecmp */

#ifndef __CKDM__
#include <stdlib.h>
#include <stdio.h>
#include <windows.h>
#endif

/* Changelog:
 * 2006-10-11: Includuju <string.h> misto "block.h"
 */

/* 
 * TODO: Osetrit ztratu vazby (potenciometr)
 */

/* Pokud je zadefinovano, rameno se pouze jednoduse simuluje.  Simulace
 * zahrnuje konecnou rychlost presunu ramena kazdopadne uz je docela
 * zastarala.*/

//#define SIMULATION

/* Parametry pohybu pro plnou a pomalou rychlost */
const struct arm_movement_params arm_movement_full = {64, 1024, 320, 2500};
const struct arm_movement_params arm_movement_slow = {64, 8, 64, 2500};

	/* Podle protokolu, uchovava priznaky ARM_BIT_* */
uint32 arm_flags;
	/* Aktualni stav ramene */
static enum arm_state arm_state;
	/* Zda byl poslan prikaz kill */
static bool arm_kill_sent;
	/* Nulovy bod ramena [celokroky] */
static int32 arm_zero;
	/* posledni znama pozice a cas ramene pro pocitani polohy[celokroky] */
static int32 arm_pos;
	/* Poloha se zmenila od posledniho uploadu */
static bool arm_pos_dirty = true;
	/* Cas kdy byl zahajen posledni pohyb (pro pocitani polohy) [milisekundy] */
static uint32 arm_time;
	/* Aktualni charakteristika pohybu ramene */
static const struct arm_movement_params *arm_movement_params;

/* Spolecne pro simulaci i nesimulaci *****************************************/

int32
arm_get_real_zero(void)
{
	return arm_zero;
}

void
arm_set_real_zero(int32 zero)
{
	arm_zero = zero;
}

/* Prevod celokroky motoru -> milistupne */
static int32
steps2mdeg(int32 mpos)
{
	return mpos * 18;
}

/* Prevod milistupne -> celokroky motoru */
static int32
mdeg2steps(int32 deg)
{
	return deg / 18;
}

/* Krokovy motor ma 1.8 stupnu na celokrok. Prevodovka je 1:100, takze rameno
 * se pohne o 0.018 stupnu na celokrok. */

/* Prevadi absolutni pozici[celokroky] na relativni [milistupne] s prihlednutim
 * k nastavene nule ramena. Dale prihlizi k tomu, ze jsem si nastavil v miste
 * praveho koncaku krok 8 000 000 (polovina maximalniho rozsahu poloh radice
 * motoru, v nejmensich mikrokrocich)). 
 *
 * POZNAMKA: tyto absolutni a relativni souradnice _nejsou_ to same jako
 * absolutni a relativni souradnice popsane v dokumentaci komunikacniho
 * protokolu!*/
static int32
pos_abs2rel(int32 pos)
{
	return steps2mdeg(pos - 8000000/64 - arm_zero);
}

/* Prevadi relativni pozici[milistupne] na absolutni [celokroky] s prihlednutim
 * k nastavene nule ramena */
static int32
pos_rel2abs(int32 pos)
{
	return mdeg2steps(pos) + arm_zero + 8000000/64;
}

/* Spocita (odhadne), o kolik se zmeni poloha ramene za dany cas.
 * t: Cas od zacatku pohybu [ms]
 * start: rychlost start/stop
 * a: zrychleni[celokroky/s^2]
 * v: max. rychlost[celokroky/s]
 * vraci: Drahu, kterou rameno urazi v case t [celokroky]
 */
static int32
arm_compute_pos(uint32 t, uint32 start, uint32 a, uint32 v)
{
	int32 _t, _s; 
	/* Cas si prevedu na desetiny sekundy, to je dostatecne rozliseni
	a zaroven zabranim pretekani ve vypoctech. */
	t /= 100;
	
	/* Cas kdy dosahnu max. rychlosti */
	_t = 10*(v-start)/a; 
	if (t < _t)
		return a*t*t/200 + start*t/10;
		
	/* draha ve ktere dosahnu max. rychlosti */
	_s = a*_t*_t/200 + start*_t/10; 
	return _s + (t-_t)*v/10;
}
	
enum cmd_err
arm_is_local(uint32 *local)
{
	*local = (arm_flags & ARM_BIT_LOCAL) ? 1 : 0;
	return cmd_ok;
}

#ifndef SIMULATION

/* Verze s HW *****************************************************************/

/* Jednim z problemu je zjisteni, zda se rameno stale jeste pohybuje. Kdyz je v
 * pohybu, nereaguje na zadne prikazy(krome K(ill) a \(reset)), nastesti aspon
 * na zaslane prikazy odpovida funkci Return (T41).  Po prikazu R(un) poslu
 * prikaz U(pload), ktery vysle hodnotu hnedka jakmile se motor zastavi. ve
 * funkci pre_cmd pak kontroluju jestli uz prisla odpoved od uploadu. Kdyz jo,
 * provedu si potrebne zalezitosti. Kdyz se vola nejaka jina funkce, pomoci
 * pre_cmd nejprve zjistim, zda uz rameno stoji, a kdyz ne, vracim cmd_wait.
 * 
 * Dalsi z problemu, ktery castecne vyplyva z predchoziho, je ze nedokazu behem
 * pohybu ramene zjistit aktualni polohu. Tohle zase resim tak, ze si pomoci
 * pre_cmd zjistim zda se rameno hybe, a kdyz ano tak si polohu spocitam z
 * vzorce kam dosadim polohu na zacatku pohybu, cas zacatku pohybu, aktualni
 * cas a parametry pohybu. Kdyz rameno stoji, vracim skutecnou polohu.
 *
 * Dalsi problem je urcovani polohy. Rameno umi pro polohu pouze kladna cisla
 * od 0 do 16 000 000 (v nejmensich mikrokrocich, tedy 1/64 kroku). Rozsah
 * ramene je zhruba 90 stupnu, prevodovka 1:100, 1.8 stupnu na krok motoru,
 * takze potrebny rozsah poloh(v krocich) je zhruba 90*100/1.8*64 = 320000
 * nejmensich mikrokroku. Mam tedy obrovskou rezervu, takze si nastavim misto,
 * kde je referencni bod (hrana praveho koncaku) na pulku max. rozsahu, tedy 
 * 8 000 000. 
 *
 * Dalsi vec, kterou musim resit, jsou koncove spinace. Rameno ma na obou
 * stranach sve vymezene polohy koncove spinace, ktere jsou jednak pripojeny na
 * vstup Limit, druhak je jeden z nich pripojen na vstup B14 (spinac v zapornem
 * smeru), a druhy z nich na vstup B15(spinac v kladnem smeru). Pozadavkem je,
 * aby se pohyb ramene pri dojezdu na koncak zastavil (zajistuje automaticky
 * vstup Limit), ale aby i v koncovych polohach bylo mozne s ramenem jeste
 * pohybovat, ovsem pouze ve smeru mezi koncaky (tj. firmware neumozni odjet
 * uplne mimo koncaky). Situaci resim tak, ze se vzdy na konci nejakeho pohybu
 * podivam, zda jsem dojel do krajni polohy a tedy sepnul nejaky spinac. Pokud
 * ano, jednak nastavim priznak ze rameno ztratilo referenci(tedy je
 * nedefinovana poloha), druhak vypnu funkci Limit, takze zacne opet byt mozne
 * s ramenem pohybovat. Pokud zjistim, ze jsem zadny koncak nesepnul, zase
 * povolim funkci Limit aby se pri narazu na koncak rameno automaticky
 * zastavilo. */

#ifdef __CKDM__
#include "cmos.h"
#endif

static enum cmd_err
rdbytes(void *buf, size_t len)
{
	int rd = ser_in_timeout(ARM_COM, buf, len, JIFRATE/4);
	if (!rd)
		return cmd_timeout;
	if (rd != len)
		return cmd_invalid;
	return cmd_ok;
}

/* vysle na seriak prikaz, tj. posle cmd a nakonec ukoncovaci znak(0x0D).
 * Predpoklada ze je zapnuta funkce Return (T41), pri ktere radic vysila zpet
 * kazdy prijaty znak pro overeni. Pokusi se prijmout zpet radicem vyslana
 * data, a zkontroluje zda jsou totozna s odeslanymi. 
 *
 * Vraci:
 * cmd_ok, pokud se vratila odpoved v poradku 
 * cmd_timeout pokud doslo k timeoutu pri cekani na data 
 * cmd_invalid pokud prijata jina data nez odeslana.
 * 
 */
static enum cmd_err
cmd_out(const char *cmd)
{
	char out_data[15], ret_data[15];
	size_t len;
	const char terminator=0x0d;
	enum cmd_err err;

	strcpy(out_data, cmd);
	len = strlen(out_data);
	out_data[len] = terminator;
	len++;
	out_data[len] = '\0';
	
	ser_out_safe(ARM_COM, out_data, len);

	if (cmd_ok == (err = rdbytes(ret_data, len))) {
		ret_data[len] = '\0';
		if (0 == strcasecmp(out_data, ret_data)) { /* prikazy case insensitive */ 
			arm_flags |= ARM_BIT_ALIVE;
			return cmd_ok;
		}
		return cmd_invalid;
	}
	arm_flags &=~ ARM_BIT_ALIVE;
	return err;
}

/* Vysle prikazy postupne ven. Pokud se vyslani nektereho z prikazu nepovedlo,
 * dalsi prikazy nevysila a vrati navratovou hodnotu posledniho vyslani.
 * Za poslednim prikazem v poli musi byt NULL.
 */
static enum cmd_err
send_commands(const char * const cmds[])
{
	const char *c;
	int i;
	enum cmd_err err;
	
	for (i = 0; (c = cmds[i]) != NULL; i++)
		if (cmd_ok != (err = cmd_out(c)))
			return err;
	return cmd_ok;
}

/* Jako cmd_out, ale posklada prikaz "%c%i"%(cmd, arg) */
static enum cmd_err
cmd_with_arg_out(char cmd, uint32 arg)
{
	char buf[11];
	char *argbegin = s_putidz(buf+1, arg, 9);
	argbegin--;
	*argbegin = cmd;
	return cmd_out(argbegin);
}

/* Prikaz upload */
static enum cmd_err
upload(byte var, byte *result)
{
	char txt[4];
	enum cmd_err err;
	
	txt[0]='U';
	if (var > 9) {
		txt[1] = var/10 + '0';
		txt[2] = var%10 + '0';
		txt[3] = '\0';
	} else {
		txt[1] = var + '0';
		txt[2] = '\0';
	}
	if (cmd_ok != (err = cmd_out(txt)))
		return err;
	return rdbytes(result, 1);
}

/* Spusti koncove stupne a posle prikaz R(un) */
static enum cmd_err
run(void)
{
	enum cmd_err err;
	const char *const commands[] = {"C7", "R", "U16", NULL};
	
	if (cmd_ok != (err = send_commands(commands)))
		arm_state = stopped;
	arm_pos_dirty = true;
	arm_time = get_jiffies();
	return err;
}

static enum cmd_err
update_range_bits(void)
{
	enum cmd_err err;
	byte inputs;

/* Verze, ktera cte koncaky primo z CKDM 
	inputs = (IO_PDR6.bit.P62 | (IO_PDR6.bit.P63 << 1));
*/
/* Vraci stav koncovych stupnu. Bit 0: sepnut koncak v zapornem smeru, bit 1:
 * sepnut koncak v kladnem smeru */
	if (
		cmd_ok != (err = cmd_out("?142")) ||
		cmd_ok != (err = upload(4, &inputs))
	   )
	   return err;
	inputs ^= 3;
	if (inputs > 3) /* to by se snad nemelo stat.. */
		err = cmd_invalid;

	if (inputs)
		arm_flags |= ARM_BIT_RANGE;
	else
		arm_flags &= ~ARM_BIT_RANGE;

	if (inputs & 1)
		arm_flags |= ARM_BIT_RANGE_NEGATIVE;
	else
		arm_flags &= ~ARM_BIT_RANGE_NEGATIVE;

	if (inputs & 2)
		arm_flags |= ARM_BIT_RANGE_POSITIVE;
	else
		arm_flags &= ~ARM_BIT_RANGE_POSITIVE;

	return err;
}

/* Predpoklada ze bylo vyslane U16, pokusi se precist odpoved a porovnat ji s
 * ARM_SERIES_NUMBER. Vraci cmd_ok, cmd_timeout nebo cmd_invalid */
static enum cmd_err
read_series_number(void)
{
	byte b;

	if (0 == ser_in_timeout(ARM_COM, &b, 1, JIFRATE/10)) {
		return cmd_timeout; 
	}
	if (b != ARM_SERIES_NUMBER) {
		return cmd_invalid;
	}
	return cmd_ok;
}


/* Kdyz uvedu rameno do pohybu, vyslu hned pote prikaz U16, ktery vraci cislo
 * vyrobni serie radice (ARM_SERIES_NUMBER). Kazda funkce, ktera potrebuje aby
 * bylo rameno zastaveno, nejprve zavola pre_cmd, ktere zkontroluje jestli uz
 * nahodou neprisla odpoved od U16. Pokud ano, prislusne ji zpracuje.
 * Vraci: 
 * cmd_ok: Rameno se zastavilo
 * cmd_comm: Chyba pri komunikaci se zarizenim (nemelo by nastat)
 * cmd_wait: Rameno stale bezi
 * cmd_invalid: Prisla neocekavana odpoved od U16.
 */
static enum cmd_err
pre_cmd(void)
{
	enum cmd_err err;
	int32 tmp;
	
	if (!(arm_flags & ARM_BIT_ALIVE)) {
		if (cmd_ok != (err = arm_init()))
			return err;
	}

	if (arm_state == stopped) {
		return cmd_ok;
	}
	if (ser_ibuflen(ARM_COM) == 0) {
		return cmd_wait;
	}
	if (cmd_ok != (err = read_series_number()))
		return err;

		/* Nyni vim ze pohyb byl prave ted ukoncen */
	if (cmd_ok != (err = cmd_out("T7"))) /* vypnu koncaky */
		return err;

	if (arm_state == referencing) {
		arm_state = stopped;
		arm_flags |= ARM_BIT_REFERENCED;
		arm_flags &= ~ARM_BIT_RANGE;
		if (cmd_ok != (err = arm_set_movement_params(&arm_movement_full)))
			return err;
		err = arm_goto(0);
		if (err == cmd_illegal_movement)
			err = cmd_ok;
		return err;
	}

	arm_state = stopped;
	arm_kill_sent = false;

	/* Ulozim aktualni polohu */
	if (cmd_ok != (err = arm_get_pos(&tmp)))
		return err;

		/* Ted si zjistim, jestli jsem nahodou nezajel do koncaku */
	if (cmd_ok != (err = update_range_bits()))
		return err;
	if (arm_flags & ARM_BIT_RANGE) {
		arm_flags &=~ ARM_BIT_REFERENCED;
		if (cmd_ok != (err = cmd_out("T42"))) /* limit off */
			return err;
	} else {
		if (cmd_ok != (err = cmd_out("C42"))) /* limit on */
			return err;
	}
	return cmd_ok;
}

/* Nasleduji implementace prikazu protokolu. ********************************/

enum cmd_err
arm_identify(const char **id)
{
	*id = "Rameno firmware " ARM_FIRMWARE_VERSION ", MeasProg s.r.o.";
	return cmd_ok;
}

enum cmd_err
arm_get_state(enum arm_state *state)
{
	*state = arm_state;
	return cmd_ok;
}

enum cmd_err
arm_get_flags(uint32 *flags)
{
	enum cmd_err err;

	switch (err = pre_cmd()) {
		case cmd_wait:
			arm_flags |= ARM_BIT_BUSY;
			break;
		case cmd_ok:
			arm_flags &= ~ARM_BIT_BUSY;
			break;
		default:
			return err;
	}
	*flags = arm_flags;
	return cmd_ok;
}

enum cmd_err
arm_set_local(uint32 local)
{
	if (local)
		arm_flags |= ARM_BIT_LOCAL;
	else
		arm_flags &= ~ARM_BIT_LOCAL;
	return cmd_ok;
}

/* Presun do zadane relativni polohy [milistupne] */
enum cmd_err
arm_goto(int32 pos)
{
	enum cmd_err err;
	enum arm_state new_arm_state;

	pos = pos_rel2abs(pos);
	if (cmd_ok != (err = pre_cmd())) 
		return err;
	new_arm_state = (pos > arm_pos) ? go_forward : go_backward;
	if ((new_arm_state == go_forward) && 
	    (arm_flags & ARM_BIT_RANGE_POSITIVE))
		return cmd_illegal_movement;
	if ((new_arm_state == go_backward) && 
	    (arm_flags & ARM_BIT_RANGE_NEGATIVE))
		return cmd_illegal_movement;
	if (cmd_ok != (err = cmd_with_arg_out('G', pos*64)))
		return err;
	arm_state = new_arm_state;
	return run();
}

enum cmd_err
arm_get_kill_sent(bool *kill_sent)
{
	enum cmd_err err;

	if (cmd_ok != (err = pre_cmd()))
		return err;
	*kill_sent = arm_kill_sent;
	return cmd_ok;
}

/* posle prikaz K(ill)
 * ten se musi pro jistotu osetrovat specialne, diky <ironic> vyjimecne
 * ortogonalni </ironic> povelove sade. Vsechny prikazy mi sice pri zapnutem
 * Return (T41) odpovidaji kopii zaslaneho povelu vcetne oddelovace prikazu
 * 0x0d, ale prikaz Kill mi odpovi pouze znakem K bez oddelovace. Ach jo... */
enum cmd_err
arm_kill(void)
{
	enum cmd_err err;
	char b;
	const char command_kill[] = {'K', 0x0d};
	
		/* Pohyb ramene mohl byt jiz ukonceny, napr. vstupem Limit */
	if (cmd_wait == (err = pre_cmd()) && !arm_kill_sent) {
		ser_out_safe(ARM_COM, command_kill, sizeof(command_kill));
		if (cmd_ok != (err = rdbytes(&b, 1)))
			return err;
		if (b != 'k' && b != 'K')
			return cmd_invalid;
		arm_kill_sent = true;
	}
	return err;
}

enum cmd_err
arm_get_pos(int32 *pos)
{
	byte b1, b2, b3;
	enum cmd_err err;
	
	err = pre_cmd();
	if (err != cmd_ok && err != cmd_wait)
		return err;

	if (arm_state == referencing) {
		*pos = pos_abs2rel(arm_pos);
		return cmd_ok; /* pri referencovani polohu vubec neznam */
	} else if (arm_state != stopped) { /* spoctu si predpokladanou pozici... */
		int32 ds = arm_compute_pos(
				(get_jiffies() - arm_time)*(1000/JIFRATE), 
				arm_movement_params->start_stop, 
				arm_movement_params->acceleration, 
				arm_movement_params->velocity);
		if (arm_state == go_negative || arm_state == go_backward) 
			ds = -ds;
		*pos = pos_abs2rel(arm_pos + ds);
		return cmd_ok;
	}
	
	if (!arm_pos_dirty) {
		*pos = pos_abs2rel(arm_pos);
		return cmd_ok;
	}
	if (
		cmd_ok != (err = upload(1, &b1)) ||
		cmd_ok != (err = upload(2, &b2)) ||
		cmd_ok != (err = upload(3, &b3))
		) {
			return err;
		}
	arm_pos = (((uint32)b1 << 0) | ((uint32)b2 << 8) | ((uint32)b3 << 16))/64;
	arm_pos_dirty = false;
	*pos = pos_abs2rel(arm_pos);
	return cmd_ok;
}

enum cmd_err
arm_set_zero(int32 zero)
{
	int32 new_zero = mdeg2steps(zero);

	if (new_zero == arm_zero)
		return cmd_ok;
	arm_zero = new_zero;
	cmos_wr();
	return cmd_ok;
}

enum cmd_err
arm_get_zero(int32 *pos)
{
	*pos = steps2mdeg(arm_zero);
	return cmd_ok;
}

/* Predpokladam, ze na vstupu B14 je koncak v zapornem smeru, na vstupu B15 je
 * koncak v kladnem smeru, a referencuji vzhledem k hrane koncaku v zapornem
 * smeru. Dale predpokladam, ze oba koncaky jsou zaroven napojeny na vstup
 * LIMIT. Predpokladam, ze kladny smer pohybu je od zaporneho koncaku ke
 * kladnemu. Predpokladam ze microsteps jsou nastaveny na 4 (M4)
 * 
 * Tato funkce meni nasledujici priznaky:
 * limit off: Vzdy na konci vypnute (C42)
 * koncaky: vzdy na konci odpojene (T7)
 * Fullsteps: vypnute (C52) */
enum cmd_err
arm_reference(void)
{
	enum cmd_err err;
#if 0
	const char *const commands[] = {
		"[",
		"C7",   /* zapnu koncaky */
		"T42",  /* limit off */
		")14",  /* odjedu z dosahu koncaku */
		"C42",  /* limit on */
		"G+",   /* Dobehnu ke kladnemu koncaku */
		"R",

		"C52",  /* Zastaveni seeku na mikrokrocich, nikoli celokrocich */
		"T42",  /* limit off */
		"S128", /* touto rychlosti (bacha na rozmezi hodnot, viz tab str. 30) */
		"(15",  /* Pomalinku seeknu na hranu. */
		"C42",  /* limit on */
		"T7",   /*vypnu koncaky */
		"=8000000",  /* 8 000 000 = 0 (referencovani polohy) */
		"U16",  /* Poslu informaci o dokonceni reference */
		"]",
		NULL
	};
#endif
	const char *const commands[] = {
		"[",
		"C7",   /* zapnu koncaky */
		"T42",  /* limit off */
		"(15",  /* odjedu z dosahu koncaku */
		"C42",  /* limit on */
		"G-",   /* Dobehnu ke kladnemu koncaku */
		"R",

		"C52",  /* Zastaveni seeku na mikrokrocich, nikoli celokrocich */
		"T42",  /* limit off */
		"S128", /* touto rychlosti (bacha na rozmezi hodnot, viz tab str. 30) */
		")14",  /* Pomalinku seeknu na hranu. */
		"C42",  /* limit on */
		"T7",   /*vypnu koncaky */
		"=8000000",  /* 8 000 000 = 0 (referencovani polohy) */
		"U16",  /* Poslu informaci o dokonceni reference */
		"]",
		NULL
	};

	if (
		cmd_ok != (err = pre_cmd()) ||
		cmd_ok != (err = arm_set_movement_params(&arm_movement_full)) ||
		cmd_ok != (err = send_commands(commands))
	   )
		return err;
	arm_state = referencing;
	return cmd_ok;
}

/* Nasleduji zbyle prikazy hlavickoveho souboru (pro praci s panelem) ******/

enum cmd_err
arm_init(void)
{
	enum cmd_err err;
	const char reset[] = {'\\', 0xd};
	const char t41[] = {'t', '4', '1', 0xd};	
	
	if (0 != ser_init(0, 4800, 0))
	   return cmd_comm;
	
	arm_flags &=~ ARM_BIT_REFERENCED;
	arm_flags &=~ ARM_BIT_ALIVE;
	arm_flags &=~ ARM_BIT_BLOCKED;

	ser_out_safe(ARM_COM, reset, sizeof(reset));
	ser_out_safe(ARM_COM, t41, sizeof(t41));
	ser_purgebuf(ARM_COM, JIFRATE/4);

	if (
		cmd_ok != (err = cmd_out("Q4")) ||
		cmd_ok != (err = cmd_out("M4")) ||
		cmd_ok != (err = cmd_out("=8000000")) ||
		cmd_ok != (err = arm_set_movement_params(&arm_movement_full))
	   )
		return err;

	arm_pos_dirty = true;
	arm_state = stopped;
	arm_kill_sent = false;
	arm_set_local(true);

	/* posilam jen abych overil ze rameno odpovida... */
	if (cmd_ok == (err = cmd_out("T41"))) 
		arm_flags |= ARM_BIT_ALIVE;
	return err;
}

#if 0 /* kdyby nahodou bylo potreba.. */
enum cmd_err
arm_moverel(int32 dist)
{
	enum cmd_err err;

	if (cmd_ok != (err = pre_cmd()))
		return err;
	if (dist < 0) {
		arm_state = go_backward;
		err = cmd_with_arg_out('B', -dist*64);
	} else {
		arm_state = go_forward;
		err = cmd_with_arg_out('F',  dist*64);
	}
	if (err != cmd_ok) {
		arm_state = stopped;
		return err;
	}
	return run();
}
#endif

/* Nekonecny pohyb dopredu */
enum cmd_err
arm_go_positive(void)
{
	enum cmd_err err;
	
	if (cmd_ok != (err = pre_cmd())) 
		return err;
	if ((arm_flags & ARM_BIT_RANGE_POSITIVE))
		return cmd_illegal_movement;
	if (cmd_ok != (err = cmd_out("G+")))
		return err;
	arm_state = go_positive;
 	return run();
}

/* Nekonecny pohyb dozadu */
enum cmd_err
arm_go_negative(void)
{
	enum cmd_err err;
	
	if (cmd_ok != (err = pre_cmd())) 
		return err;
	if ((arm_flags & ARM_BIT_RANGE_NEGATIVE))
		return cmd_illegal_movement;
	if (cmd_ok != (err = cmd_out("G-")))
		return err;
	arm_state = go_negative;
 	return run(); 	
}

/* Nastavi rychlost (zrychleni, start/stop, profile...) pohybu ramene. */
enum cmd_err
arm_set_movement_params(const struct arm_movement_params *params)
{
	enum cmd_err err;
	if (
		cmd_ok != (err = pre_cmd()) ||
		cmd_ok != (err = cmd_out("[")) ||
		cmd_ok != (err = cmd_with_arg_out('S', params->start_stop)) ||
		cmd_ok != (err = cmd_with_arg_out('V', params->velocity)) ||
		cmd_ok != (err = cmd_with_arg_out('A', params->acceleration)) ||
		cmd_ok != (err = cmd_with_arg_out('P', params->profile)) ||
		cmd_ok != (err = cmd_out("U16")) ||
		cmd_ok != (err = cmd_out("]")) ||
		cmd_ok != (err = read_series_number())
	   )
		return err;
	arm_movement_params = params;
	return cmd_ok;
}

#else /* SIMULACE */

/* Simulace *******************************************************************/

#include "timer.h"
#include "cmos.h"

/* v jednom kroku je 64 mikrokroku */

/* [mikrokroky/s] */
static uint32 arm_speed;
/* Absolutni poloha (doraz = 1) [mikrokroky] */
static uint32 arm_pos; 
/* vzdalenost kterou mam jet [mikrokroky] */
static uint32 arm_dist;

/* Prevod nejmensi mikrokroky motoru -> milistupne */
static int32
msteps2mdeg(int32 mpos)
{
	return (mpos / 64) * 18;
}

/* Prevod milistupne -> nejmensi mikrokroky motoru */
static int32
mdeg2msteps(int32 deg)
{
	return (deg / 18) * 64;
}

enum cmd_err
arm_init(void)
{
	arm_pos = 64;
}

enum cmd_err
arm_go_positive(void)
{
	arm_state = go_positive;
	return cmd_ok;
}

enum cmd_err
arm_go_negative(void)
{
	arm_state = go_negative;
	return cmd_ok;
}

enum cmd_err
arm_kill(void)
{
	arm_state = stopped;
	return cmd_ok;
}

enum cmd_err
arm_velocity(uint32 v)
{
	arm_speed = mdeg2msteps(v);
	return cmd_ok;
}
	
enum cmd_err
arm_get_pos(int32 *pos)
{
	dword tmp = jiffies;
	uint32 diff = arm_speed * (tmp - arm_time) / JIFRATE;
	arm_time = tmp;
	
	switch(arm_get_state()) {
		case go_forward: case go_backward:
			if (diff > arm_dist)
				diff = arm_dist;
	}
	arm_dist -= diff;
	switch(arm_get_state()) {
		case go_forward: case go_positive:
			arm_pos += diff;
			break;
		case go_backward: case go_negative:
			arm_pos -= diff;
			break;
	}
	switch(arm_get_state()) {
		case go_forward: case go_backward:
			if (arm_dist <= 0)
				arm_state = stopped;
	}
	*pos = msteps2mdeg((int32)arm_pos - arm_zero);
	return cmd_ok;
}

enum cmd_err
arm_goto(int32 where)
{
	enum cmd_err e;
	int32 mpos = msteps2mdeg(arm_pos);
	if (cmd_ok != (e = arm_velocity(VELOCITY_FAST)))
		return e;
	where += msteps2mdeg(arm_zero);
	return arm_moverel(where - mpos);
}

enum cmd_err
arm_reference(void)
{
	return arm_goto(90000 - msteps2mdeg(arm_zero));
}

enum cmd_err
arm_set_zero(int32 where)
{
	arm_zero = mdeg2msteps(where);
	cmos_wr();
	return cmd_ok;
}

enum cmd_err
arm_get_zero(int32 *zero)
{
	*zero = msteps2mdeg(arm_zero);
	return cmd_ok;
}

#endif /* SIMULATION */
