#include "luxmetr.h"
#include "panel.h"
#include "timer.h"
#include "ui.h"
#include "gpib.h"
#include "misc.h"
#include <string.h>

#define ERR_COMM -1
#define ERR_INTR -2
#define ERR_LIGHT -3
#define ERR_VARLIGHT -4

#define LUX_DEFAULT_RANGE 5

#define GPIB_PRE {\
	byte gpib_tries; \
	for (gpib_tries = 0; gpib_tries < 3; gpib_tries++) {

#define GPIB_POST \
		if (gpib_get_status() == GPIB_NONE) { \
			lux_flags |= LUX_BIT_ALIVE; \
			return cmd_ok; \
		} else gpib_reset(); \
	}\
	lux_flags &=~ LUX_BIT_ALIVE; \
	return cmd_comm; \
}

/* Po zmene rozsahu je dobre chvili pockat, nez se hodnota trochu ustali. k
 * tomu pouzivam promennou lux_read_time. */
static uint32 lux_read_time = 0;
/* Luxmetr ma rozsahy <0, 6> */
static signed char lux_range;
static uint32 lux_flags;
static int32 lux_val;

static void
set_state_measuring(void)
{
	lux_flags |=  LUX_BIT_MEASURING;
	lux_flags &=~ LUX_BIT_MEASURED;
}

static void
set_state_measured(void)
{
	lux_flags &=~ LUX_BIT_MEASURING;
	lux_flags |=  LUX_BIT_MEASURED;
}

static void
set_state_failed(void)
{
	lux_flags &=~ LUX_BIT_MEASURING;
	lux_flags &=~ LUX_BIT_MEASURED;
}

#define set_state_idle set_state_failed

/* Neukladam do int32, protoze obsahuje jako informaci navic rozsah pri kterem
 * probehlo mereni. */
struct lux_val {
	int val;
	byte exp;
};

static enum cmd_err
lux_do_set_range(char range)
{
	char cmd[] = "Rx";
	/* Rozsah luxmetru se nastavuje prikazem R[0-6] */
	GPIB_PRE

	cmd[1] = range + '0';
	gpib_wrstr(LUX_ADDR, cmd);

	GPIB_POST
}

/* Nastavi rozsah luxmetru, luxmetr ma rozsahy <0,6> */
enum cmd_err /* XXX static */
lux_set_range(char range)
{
	enum cmd_err err = lux_do_set_range(range);
	if (err == cmd_ok) {
		lux_range = range;
		/* Chvilku trva nez se hodnota na displeji ustali.. */
		lux_read_time = jiffies + JIFRATE/2;
		lux_flags &=~ LUX_BIT_LOCAL;
	}
	return err;
}

/* Uvede luxmetr do lokalniho rezimu. Predpoklada ze jsem v (RWLS) */
static enum cmd_err
lux_set_local_true(void)
{
	GPIB_PRE

	gpib_gtl(LUX_ADDR);
	if (gpib_get_status() == GPIB_NONE)
		lux_flags |= LUX_BIT_LOCAL;

	GPIB_POST
}

/* Zakaze lokalni rezim luxmetru (kolecko pro zmenu rozsahu). Predpoklada ze
 * jsem v (LWLS).  */
static enum cmd_err
lux_set_local_false(void)
{
	GPIB_PRE

	gpib_mim(MIM_MLA(LUX_ADDR));
	if (gpib_get_status() == GPIB_NONE)
		lux_flags &=~ LUX_BIT_LOCAL;

	GPIB_POST
}

/* Tohle skutecne pouze precte displej. */
static enum cmd_err
lux_do_rddata(char *buf, byte maxlen)
{
	byte cnt;
	while(jiffies < lux_read_time)
		go_sleep();

	GPIB_PRE

	gpib_get(LUX_ADDR);
	cnt = gpib_read(LUX_ADDR, buf, maxlen);
	buf[cnt] = '\0';

	GPIB_POST
}

/* Prijme zpravu z displeje luxmetru.
 * Zprava je ve formatu x.xxxEx, kde jednotlive cislice jsou <= 1.999E6.  Pokud
 * nastalo preteceni, nakopiruje to tam retezec OL (overflow). Pokud jsem
 * precetl neco jineho, povazuji to za chybu a zkusim cist jeste jednou, pokud
 * se nepovede ani to, vracim cmd_invalid.
 *
 * Pokud je na fyzickem dispeji zaporne cislo (i to se muze stat), prectu ho
 * bez znamenka. Je to nicmene dost nestandardni situace na to abych se s tim
 * netrapil.
 */
enum cmd_err /* XXX static */
lux_rddata(char *buf, byte maxlen)
{
	enum cmd_err err;
	byte tries;

	for (tries = 0; tries < 2; tries++) {
		if (cmd_ok != (err = lux_do_rddata(buf, maxlen)))
			return err;
		if (buf[0] == 'O' && buf[1] == 'L')
			return cmd_ok;
		if (
			isnum(buf[0]) && 
			'.' ==buf[1]  && 
			isnum(buf[2]) && 
			isnum(buf[3]) && 
			isnum(buf[4]) && 
			'E' ==buf[5]  && 
			isnum(buf[6])
		) {
			lux_flags &=~ LUX_BIT_LOCAL;
			return cmd_ok;
		}
	}
	return cmd_invalid;
}



/* Precte hodnotu z luxmetru. Pokud je na displeji preteceni, hodi do val
 * straslive velkou hodnotu.
 * Luxmetr vraci 1.999E6 (cislice udavaji max. hodnotu prislusneho mista)
 * A pak dalsi, TODO prozkoumat - zaporne cislo, chyba */
static enum cmd_err 
lux_rddisplay(struct lux_val *val)
{
	char buf[15];
	enum cmd_err err;

	if (cmd_ok != (err = lux_rddata(buf, sizeof(buf))))
		return err;

	if (buf[1] != '.') {
		val->val = 21474; /* 2^31(=maxint32) = 2147483648 > 21474 * 10^5 */
		val->exp = 5;
	} else {
		val->val = (buf[0]-'0')*1000 + (buf[2]-'0')*100 + (buf[3]-'0')*10 + (buf[4]-'0');
		val->exp = buf[6] - '0';
		lux_range = val->exp;
	}
	return cmd_ok;
}

/* 
 * Vrati jednu hodnotu, sam si nastavi rozsahy tak aby mereni nepretekalo ani
 * nepodtekalo. Je implementovana hystereze, o rozsah nize se prepnu pokud
 * hodnota < LUX_UNDERFLOW_VAL/1000, o rozsah vyse pokud je hodnota >
 * LUX_OVERFLOW_VAL/1000 (max. je 1.999).
 */
static enum cmd_err
lux_rdval(struct lux_val *val, bool (* callback)(void))
{
	byte tries = 0;
	
		/* zpravu ctu ve tvaru [01].xxxEr (xxx * 10^(r-3)) */
	while(tries++ < 10) {
		if (callback && callback()) {
			val->val = ERR_INTR;
			goto failure;
		}
		if (cmd_ok != lux_rddisplay(val)) {
			val->val = ERR_COMM;
			goto failure;
		}

		if (val->val > LUX_OVERFLOW_VAL) {	/* preteceni */
			if (lux_range >= 6) {
				val->val = ERR_LIGHT;
				goto failure;
			}
			if (cmd_ok != lux_set_range(lux_range + 1)) {
				val->val = ERR_COMM;
				goto failure;
			}
			continue;
		}
		if (val->val < LUX_UNDERFLOW_VAL && lux_range > 0) {	/* podteceni */
			if (cmd_ok != lux_set_range(lux_range - 1)) {
				val->val = ERR_COMM;
				goto failure;
			}
			continue;
		}
		/* hodnota ok */
		return cmd_ok;
	}
	/* Furt kolisa rozsah, to se musi hodne menit osvetleni... */
	val->val = ERR_VARLIGHT;
failure:
	set_state_failed();
	return cmd_comm;
}

static bool
in_tolerance(struct lux_val *a, struct lux_val *b)
{
	const int32 digerr = 3;
	return a->exp == b->exp && ABS(a->val - b->val) <= digerr;
}

/* Nasleduji implementace prikazu podle protokolu ***************************/

enum cmd_err
lux_init(void)
{
	set_state_idle();
	lux_flags = 0;
	gpib_reset();
	gpib_dcl(LUX_ADDR);
	lux_flags &=~ LUX_BIT_LOCAL;
	return lux_set_range(LUX_DEFAULT_RANGE);
}

enum cmd_err
lux_set_local(bool local)
{
	if (local)
		return lux_set_local_true();
	else
		return lux_set_local_false();
}

enum cmd_err
lux_get_flags(uint32 *flags)
{
	*flags = lux_flags;
	return cmd_ok;
}

enum cmd_err
lux_measure(bool (* callback)(void))
{
	int i;
	struct lux_val window[LUX_WINDOW_WIDTH];
	struct lux_val *val;
	int tries = 0;
	
	if (cmd_ok != lux_set_local(false)) {
		lux_val = (int32)ERR_COMM;
		goto failure;
	}
	set_state_measuring();
	for (i = 0; i < LUX_WINDOW_WIDTH; i++) {
		val = window + i;
		if (cmd_ok != lux_rdval(val, callback)) {
			lux_val = (int32)ERR_COMM;
			goto failure;
		}
		if (val->val < 0) {
			lux_val = (int32)(val->val);
			goto failure;
		}
	}
	while(tries++ < 100) {
		if (callback && callback()) {
			lux_val = (int32)ERR_INTR;
			goto failure;
		}
		for (i = 1; i < LUX_WINDOW_WIDTH - 1; i++) 
			if (!in_tolerance(window + i, window + LUX_WINDOW_WIDTH - 1))
				goto shift;
		/* Zprumeruju, exponent vezmu z posledniho mereni, diky in_tolerance
		 * mam zaruceno ze vsechny exponenty jsou stejne. */
		lux_val = 0;
		for (i = 0; i < LUX_WINDOW_WIDTH; i++)
			lux_val += (int32)(window[i].val);
		lux_val /= (int32)LUX_WINDOW_WIDTH;
		for (i = window[0].exp; i; i--)
			lux_val *= (int32)10;
		set_state_measured();
		return cmd_ok;
shift:
		memmove(window, window + 1, sizeof(struct lux_val) * (LUX_WINDOW_WIDTH-1));
		val = window + LUX_WINDOW_WIDTH - 1;
		lux_rdval(val, callback);
		if (val->val < 0) {
			lux_val = val->val;
			goto failure;
		}
	}

	lux_val = ERR_VARLIGHT;

failure:
	set_state_failed();
	return cmd_ok;
}

enum cmd_err
lux_get_val(int32 *val)
{
	*val = lux_val;
	return cmd_ok;
}
