#include "globals.h"
#include "panel.h"
#include "rtc.h"
#include "sleep.h"

/* Changelog:
 *
 * 2006-11-16: Rozdeleni na praci s datocasem a CMOS
 */

#define WRCn IO_PDR5.bit.P50             
#define SRTCn IO_PDR1.bit.P14
#define SCKn IO_PDR4.bit.P47
#define SIN IO_PDR4.bit.P46
#define SOUT IO_PDR5.bit.P57

#define mode_WRCn IO_DDR5.bit.D50
#define mode_SRTCn IO_DDR1.bit.D14
#define mode_SCKn IO_DDR4.bit.D47
#define mode_SIN IO_DDR4.bit.D46
#define mode_SOUT IO_DDR5.bit.D57

void
rtc_init(void)
{
	mode_SIN=1;
	mode_WRCn=1;
	mode_SRTCn=1;
	mode_SCKn=1;
	mode_SOUT=0;
};

/* Pomocne funkce **********************************************************/

/* A=<hodnota><adresa> */
static byte 
rtcinout(byte a)
{
	byte w,x;

	SRTCn = 0;
	for(w = 8; w; w--){
		for(x = 30; x; x--) 
			__wait_nop();
		SCKn = 0;
		SIN = a & 1;
		a >>= 1;
		a |= SOUT << 7;
		for(x = 30; x; x--)
			__wait_nop();
		SCKn = 1;
	};
	for (x = 30; x; x--)
		__wait_nop();
	SRTCn = 1;
	for(x = 30; x; x--)
		__wait_nop();
	return a>>= 4;
};

/* A=<hodnota><adresa> */
static byte 
wrrtca(byte a)
{
	WRCn=0;
	return rtcinout(a);
};

static byte 
rdrtc(byte a)
{
	WRCn=1;
	return rtcinout(a);
};

/* Ceka az bude BUSY flag shozen, pak mame nejake 3.8 ms casu na operace s RTC.
 * vraci 0 -- OK, jinak -- Nejaka hrozna chyba. */
static byte
rtc_wait_busy(void)
{
	long x;
	
	rdrtc(14);
	for (x = 65536; x && rdrtc(14) & 8; x--);
	return x == 0;
}

/* Funkce pro praci s RTC/CMOS *********************************************/

#ifdef USE_RTC_CMOS

/**
 * rtc_wr4
 * Zapise spodni 4 bity val do CMOS RTCcka
 * @param addr adresa kam se ma psat v rozsahu <0, 29>
 * @param val spodni 4 bity jsou data k zapsani
 */
static void 
rtc_wr4(byte addr, byte val)
{
	__DI();
	if(rtc_wait_busy()) {
		__EI(); 
		return;
	}

	if (addr > 14) {
		wrrtca(0x3f); /* Vyber mod */
		addr -= 15;
	} else
		wrrtca(0x2f); /* Vyber mod */

	wrrtca((val << 4) | addr);
	wrrtca(0x0f); /* zpatky do modu 0 */
	__EI();
}

/**
 * rtc_rd4
 * Vrati 4 bity z CMOS RTCcka
 * @param addr adresa v rozsahu <0, 29>
 * @return spodni ctyri bity vysledku obsahuji data z CMOS
 */
static byte 
rtc_rd4(byte addr)
{
	byte retval;
	__DI();
	
	wrrtca(0x0f);
	if(rtc_wait_busy()) {
		__EI(); 
		return -1;
	}

	if (addr > 14) {
		wrrtca(0x3f); /* Vyber mod */
		addr -= 15;
	} else
		wrrtca(0x2f); /* Vyber mod */

	rdrtc(addr); /* 'Precti' */
	retval = rdrtc(0); /* Ctena hodnota se objevi az po jednom cteni !! */
	wrrtca(0x0f); /* zpatky do modu 0 */
	__EI();
	return retval;
}

void
cmos_rdnibble(byte addr, byte *val, byte *chksum)
{
	*val = rtc_rd4(addr);
	if (chksum)
		*chksum += *val;
}

void
cmos_rdbyte(byte addr, byte *val, byte *chksum)
{
	byte tmp;
	cmos_rdnibble(addr, &tmp, chksum);
	*val = tmp;
	cmos_rdnibble(addr + 1, &tmp, chksum);
	*val |= (tmp << 4);
}

void
cmos_rdword(byte addr, word *val, byte *chksum)
{
	byte tmp;
	cmos_rdbyte(addr, &tmp, chksum);
	*val = tmp;
	cmos_rdbyte(addr + 2, &tmp, chksum);
	*val |= ((word)tmp << 8);
}

void
cmos_rddword(byte addr, dword *val, byte *chksum)
{
	word tmp;
	cmos_rdword(addr, &tmp, chksum);
	*val = tmp;
	cmos_rdword(addr + 4, &tmp, chksum);
	*val |= ((dword)tmp << 16);
}

/* false -- ok, true -- chyba */
bool
cmos_wrnibble(byte addr, byte val, byte *chksum)
{
	byte tmp;

	val &= 0xF;
	rtc_wr4(addr, val);

	cmos_rdnibble(addr, &tmp, NULL);
	if (tmp != val)
		return true;
	
	if (chksum)
		*chksum += val;
	return false;
}

/* false -- ok, true -- chyba */
bool
cmos_wrbyte(byte addr, byte val, byte *chksum)
{
	if (
			cmos_wrnibble(addr, val & 0xF, chksum) ||
			cmos_wrnibble(addr + 1, val >> 4, chksum)
	   )
		return true;
	else
		return false;
}

/* false -- ok, true -- chyba */
bool
cmos_wrword(byte addr, word val, byte *chksum)
{
	if (
			cmos_wrbyte(addr, val & 0xFF, chksum) ||
			cmos_wrbyte(addr + 2, val >> 8, chksum)
	   )
		return true;
	else
		return false;
}

/* false -- ok, true -- chyba */
bool
cmos_wrdword(byte addr, dword val, byte *chksum)
{
	if (
			cmos_wrword(addr, val & 0xFFFF, chksum) ||
			cmos_wrword(addr + 4, val >> 16, chksum)
	   )
		return true;
	else
		return false;
}

#endif /* USE_RTC_CMOS */

/* Funkce pro cteni/nastavovani data/casu **********************************/

#ifdef USE_RTC_DATETIME

void
set_string_datetime(const char *dt)
{	
#define X(pos,nibble) (dword)(dt[pos] - '0')<<(4*nibble)
	dword datum = X(0,1)|X(1,0)|X(3,3)|X(4,2)|X(6,7)|X(7,6)|X(8,5)|X(9,4);
	dword cas = X(11,7)|X(12,6)|X(14,5)|X(15,4);
	rtc_set_datetime(datum, cas);
#undef X
}

void
get_string_datetime(char *dt)
{
	dword tmp;

	dt[2] = dt[5] = '.'; dt[10] = ' ';dt[13] = ':';dt[16] = '\0';
	
#define X(pos, nibble) (dt[pos] = ((tmp>>(nibble*4))&15) + '0')
	tmp = rtcdate();
	X(0,1);X(1,0);X(3,3);X(4,2);X(6,7);X(7,6);X(8,5);X(9,4);
	tmp = rtctime();
	X(11,7);X(12,6);X(14,5);X(15,4);
#undef X
}

/* Inkrementuje citac na adrese b e <0, 12> c-krat. */
static void 
setdigb(byte b,byte c)
{
	b&=15;
	for(c&=31;c;c--)
		wrrtca(b);
};

byte 
rtc_set_datetime(dword datum,dword cas)
{
	long x;
	/* kdysi davno tu bylo __DI a __EI, elsaco rutiny pro praci s rtcckem asi
	 * nechtely byt ruseny inty, nevim jestli to vadi nebo ne.  protoze to ale
	 * cely trva dost dlouho, neumerne to prodluzuje pipnuti OK, takze se
	 * namaskuju do 6ky aby mi fungoval aspon buzzer_int kterej stejne trva
	 * chvilku... :-)
	 * ostatni nedulezity preruseni ale stejne zakazuju. */

	__set_il(6);
	
	wrrtca(0x8f);/* prazdny zapisovy cyklus */
	wrrtca(0x8f);/* RESET */
	wrrtca(0x0f);/* select mode */
	if(rtc_wait_busy()) {
		__set_il(7);
		return 1;
	};
	wrrtca(0x1d);/* 24hour mode */
	setdigb(12,datum>>20&15);/* desitky let (0) */
	setdigb(11,datum>>16&15);/* jednotky let (0) */
	setdigb(9,(datum>>8&15)+(datum&0xf000?10:0)-1);/* mesice (01) */
	x=(datum&15)+__mulu(datum>>4&15,10)-1;
	setdigb(8,__divu(x,10));/* desitky dni (0) */
	setdigb(7,__modu(x,10));/* jednotky dni (1) */
	setdigb(6,cas&7);/* den v tydnu */
	setdigb(4,(cas>>24&15)+__mulu(cas>>28&3,10));/* hodiny (00 nebo 12 ?) */
	setdigb(3,cas>>20&7);/* desitky minut (0) */
	setdigb(2,cas>>16&15);/* jednotky minut (0) */
	setdigb(1,cas>>12&7);/* desitky sekund (0) */
	setdigb(0,cas>>8&15);/* jednotky sekund (0) */
	__set_il(7);
	return 0;
};

long 
rtcdate(void)
{
	byte b;
	long x;

	__set_il(6);
	wrrtca(0x0f); /* select mode */
	if(rtc_wait_busy()) {
		__set_il(7);
		__EI(); 
		return 0;
	}
	x=0x20;
	rdrtc(b=12);
	do{
		b--;
		x<<=4;
		x|=rdrtc(b);
	}while(b>6);
	__set_il(7);
	return x;
};

long 
rtctime(void)
{
	byte b;
	long x = 0;

	__set_il(6);
	wrrtca(0x0f); /* select mode */
	if(rtc_wait_busy()) {
		__set_il(7);
		return 0;
	}
	rdrtc(b=6);
	do{
		b--;
		x<<=4;
		x|=rdrtc(b);
	}while(b<255);
	b=x>>24;
	x<<=8;
	x|=b;
	x&=0x7fffff0f;
	__set_il(7);
	return x;
};

#endif /* USE_RTC_DATETIME */
